==> 0.1.1.zip: mqtt-exporter-rs/src/main.rs <==
use backoff::backoff::Backoff;
use config::Config;
use core::str;
use json::JsonValue;
use log::{debug, error, info, warn};
use metrics::gauge;
use metrics_exporter_prometheus::PrometheusBuilder;
use metrics_util::MetricKindMask;
use regex::Regex;
use rumqttc::{
    AsyncClient, ConnAck,
    Event::{self},
    MqttOptions, Packet, Publish, QoS, SubscribeFilter,
};
use serde::Deserialize;
use std::{
    net::{IpAddr, Ipv6Addr},
    process::exit,
    time::{Duration, SystemTime, UNIX_EPOCH},
};

use mqtt_exporter::transforms::*;

struct ConfigDefaults;
impl ConfigDefaults {
    fn mqtt_address() -> String {
        "localhost".to_owned()
    }
    fn mqtt_port() -> u16 {
        1883
    }
    fn mqtt_topic() -> Vec<String> {
        vec!["#".to_owned()]
    }
    fn mqtt_keepalive() -> u64 {
        60
    }
    fn mqtt_client_id() -> String {
        "mqtt-exporter-rs".to_owned()
    }
    fn mqtt_expose_client_id() -> bool {
        true
    }
    fn expose_last_seen() -> bool {
        false
    }
    fn prometheus_address() -> IpAddr {
        IpAddr::V6(Ipv6Addr::from_bits(0))
    }
    fn prometheus_port() -> u16 {
        9000
    }
    fn prometheus_prefix() -> String {
        "mqtt".to_owned()
    }
    fn topic_label() -> String {
        "topic".to_owned()
    }
    fn hubitat_topic_prefixes() -> Vec<String> {
        vec!["hubitat".to_owned()]
    }
    fn zigbee2mqtt_availability() -> bool {
        false
    }
    fn zwave_topic_prefix() -> String {
        "zwave/".to_owned()
    }
}

#[derive(Debug, Deserialize)]
struct AppConfig {
    #[serde(default = "ConfigDefaults::mqtt_address")]
    mqtt_address: String,
    #[serde(default = "ConfigDefaults::mqtt_port")]
    mqtt_port: u16,
    #[serde(default = "ConfigDefaults::mqtt_topic")]
    mqtt_topic: Vec<String>,
    #[serde(default = "ConfigDefaults::mqtt_keepalive")]
    mqtt_keepalive: u64,
    #[serde(default)]
    mqtt_username: Option<String>,
    #[serde(default)]
    mqtt_password: Option<String>,
    #[serde(default = "ConfigDefaults::mqtt_client_id")]
    mqtt_client_id: String,
    #[serde(default = "ConfigDefaults::mqtt_expose_client_id")]
    mqtt_expose_client_id: bool,
    #[serde(default = "ConfigDefaults::expose_last_seen")]
    expose_last_seen: bool,
    #[serde(default)]
    mqtt_ignored_topics: Vec<String>,
    #[serde(with = "serde_regex", default)]
    mqtt_ignored_topics_re: Option<Regex>,
    #[serde(default = "ConfigDefaults::prometheus_address")]
    prometheus_address: IpAddr,
    #[serde(default = "ConfigDefaults::prometheus_port")]
    prometheus_port: u16,
    #[serde(default = "ConfigDefaults::prometheus_prefix")]
    prometheus_prefix: String,
    #[serde(default = "ConfigDefaults::topic_label")]
    topic_label: String,
    #[serde(default)]
    esphome_topic_prefixes: Vec<String>,
    #[serde(default = "ConfigDefaults::hubitat_topic_prefixes")]
    hubitat_topic_prefixes: Vec<String>,
    #[serde(default = "ConfigDefaults::zigbee2mqtt_availability")]
    zigbee2mqtt_availability: bool,
    #[serde(default = "ConfigDefaults::zwave_topic_prefix")]
    zwave_topic_prefix: String,
}
struct App {
    config: AppConfig,
    transformers: Vec<Box<dyn Transform>>,
}

fn parse_number(v: &str) -> Option<f64> {
    if let Ok(num) = v.parse::<f64>() {
        Some(num)
    } else if v.eq_ignore_ascii_case("on")
        || v.eq_ignore_ascii_case("true")
        || v.eq_ignore_ascii_case("online")
    {
        Some(1.0)
    } else if v.eq_ignore_ascii_case("off")
        || v.eq_ignore_ascii_case("false")
        || v.eq_ignore_ascii_case("offline")
    {
        Some(0.0)
    } else {
        None
    }
}

impl App {
    fn parse_metric(value: &JsonValue) -> Option<f64> {
        match value {
            JsonValue::Null => None,
            JsonValue::Boolean(v) => {
                if *v {
                    Some(1.0)
                } else {
                    Some(0.0)
                }
            }
            JsonValue::Number(v) => Some((*v).into()),
            JsonValue::Short(v) => parse_number(v.as_str()),
            JsonValue::String(v) => parse_number(v.as_str()),
            _ => None,
        }
    }

    fn parse_metrics(
        &self,
        mut data: JsonValue,
        topic: &str,
        _original_topic: &str,
        metric_path: Vec<String>,
        labels: &Vec<(String, String)>,
    ) {
        match data {
            JsonValue::Array(_) => {
                for (metric, child) in data.members_mut().enumerate() {
                    let mut metric_path = metric_path.clone();
                    metric_path.push(metric.to_string());
                    self.parse_metrics(child.take(), topic, _original_topic, metric_path, labels)
                }
            }
            JsonValue::Object(_) => {
                for (metric, child) in data.entries_mut() {
                    let mut metric_path = metric_path.clone();
                    metric_path.push(metric.to_string());
                    self.parse_metrics(child.take(), topic, _original_topic, metric_path, labels)
                }
            }
            _ => {
                if let Some(v) = Self::parse_metric(&data) {
                    let metric_s = metric_path.join("_");
                    let mut labels = labels.clone();
                    labels.push((self.config.topic_label.to_owned(), topic.to_owned()));
                    debug!(target: "mqtt-exporter::prom", "metric: {metric_s} value: {v} labels: {labels:?}");

                    if self.config.expose_last_seen {
                        let metric_ts = format!("{}_ts", metric_s);
                        let ts_gauge = gauge!(metric_ts, &labels);
                        let ts = SystemTime::now().duration_since(UNIX_EPOCH).unwrap();
                        ts_gauge.set(ts)
                    }
                    let gauge = gauge!(metric_s, &labels);
                    gauge.set(v);
                }
            }
        }
    }

    fn topic_filter(&self, topic: &str) -> bool {
        !self
            .config
            .mqtt_ignored_topics
            .iter()
            .any(|pfx| topic.starts_with(pfx))
            && !self
                .config
                .mqtt_ignored_topics_re
                .as_ref()
                .is_some_and(|re| re.is_match(topic))
    }

    fn deserialize_payload(payload: &[u8]) -> Result<JsonValue, Box<dyn std::error::Error>> {
        Ok(json::parse(str::from_utf8(payload)?)?)
    }

    fn handle_event(&self, event: &Event) {
        match event {
            Event::Incoming(Packet::Publish(notification)) => self.handle_publish(notification),
            Event::Incoming(Packet::ConnAck(ack)) => self.handle_connack(ack),
            // Event::Incoming(Packet::SubAck(ack)) => self.handle_suback(ack),
            Event::Outgoing(_) => {}
            e => debug!(target: "mqtt-exporter::mqtt", "Unhandled event: {:?}", e),
        }
    }

    fn handle_publish(&self, event: &Publish) {
        match Self::deserialize_payload(&event.payload) {
            Ok(parsed) => self.update_metrics(&event.topic, parsed),
            Err(e) => {
                error!(target: "mqtt-exporter::mqtt", "Error deserializing JSON payload: {:?}", e)
            }
        }
    }

    fn handle_connack(&self, ack: &ConnAck) {
        match ack.code {
            rumqttc::ConnectReturnCode::Success => {
                info!(target: "mqtt-exporter::mqtt", "Successfully connected to MQTT broker")
            }
            code => {
                warn!(target: "mqtt-exporter::mqtt", "Non success connection return code {:?}", code)
            }
        }
    }

    fn update_metrics(&self, raw_topic: &str, mut data: JsonValue) {
        let mut labels = Vec::new();

        if !&self.topic_filter(raw_topic) {
            return;
        }
        debug!(target: "mqtt-exporter::mqtt", "topic {} payload {}", raw_topic, data.pretty(2));
        let mut topic = raw_topic.to_owned();

        (topic, data) = self
            .transformers
            .iter()
            .filter(|txfm| txfm.applies_to(raw_topic))
            .fold((topic, data), |(topic, data), txfm| {
                txfm.transform(&topic, data, &mut labels)
            });

        // Should be a transform itself?
        // let topic = topic.replace('/', "_");

        let metric_path = vec![self.config.prometheus_prefix.clone()];
        self.parse_metrics(data, &topic, &topic, metric_path, &labels);
    }

    async fn subscribe(&self, client: &AsyncClient) {
        match client
            .subscribe_many(self.config.mqtt_topic.iter().map(|topic| SubscribeFilter {
                path: topic.to_owned(),
                qos: QoS::AtLeastOnce,
            }))
            .await
        {
            Ok(_) => {
                info!(target: "mqtt-exporter::mqtt", "Successfully subscribed to topics `{}`", &self.config.mqtt_topic.join(" "))
            }
            Err(e) => {
                error!(target: "mqtt-exporter::mqtt", "Failed to subscribe to topics `{}` - {}", &self.config.mqtt_topic.join(" "), e)
            }
        }
    }

    async fn mqtt_task(self) {
        let mut opts = MqttOptions::new(
            &self.config.mqtt_client_id,
            &self.config.mqtt_address,
            self.config.mqtt_port,
        );
        opts.set_keep_alive(Duration::from_secs(self.config.mqtt_keepalive))
            .set_max_packet_size(150000, 150000);

        if let (Some(username), Some(password)) =
            (&self.config.mqtt_username, &self.config.mqtt_password)
        {
            opts.set_credentials(username, password);
        }

        let (client, mut eventloop) = AsyncClient::new(opts, 10);

        self.subscribe(&client).await;

        let mut bo = backoff::ExponentialBackoff::default();

        loop {
            match eventloop.poll().await {
                Ok(event) => {
                    bo.reset();
                    self.handle_event(&event)
                }
                Err(e) => {
                    error!(target: "mqtt-exporter::mqtt", "Connection error: {}", e);
                    let delay = bo.next_backoff();
                    match delay {
                        Some(delay) => {
                            debug!(target: "mqtt-exporter::mqtt", "Backing off for {}s", delay.as_secs_f32());
                            tokio::time::sleep(delay).await
                        }
                        None => {
                            error!(target: "mqtt-exporter::mqtt", "Connection timed out, giving up");
                            exit(-1)
                        }
                    }
                }
            }
        }
    }
    async fn prom_task(&self) {
        let builder = PrometheusBuilder::new()
            .with_http_listener((self.config.prometheus_address, self.config.prometheus_port))
            .idle_timeout(MetricKindMask::ALL, Some(Duration::from_secs(300)));
        let builder = if self.config.mqtt_expose_client_id {
            builder.add_global_label("client_id", &self.config.mqtt_client_id)
        } else {
            builder
        };
        builder.install().unwrap();
    }

    fn new(config: AppConfig) -> Self {
        let mut transformers: Vec<Box<dyn Transform>> = Vec::new();
        if !config.zwave_topic_prefix.is_empty() {
            transformers.push(Box::new(ZwaveNormalize::new(&config.zwave_topic_prefix)))
        }
        if !config.esphome_topic_prefixes.is_empty() {
            transformers.push(Box::new(EsphomeNormalize::new(
                &config.esphome_topic_prefixes,
            )))
        }
        if !config.hubitat_topic_prefixes.is_empty() && !config.hubitat_topic_prefixes[0].is_empty() {
            transformers.push(Box::new(HubitatNormalize::new(
                &config.hubitat_topic_prefixes,
            )))
        }
        if config.zigbee2mqtt_availability {
            transformers.push(Box::new(Zigbee2MqttAvailability::default()))
        }
        debug!(target: "mqtt-exporter::config", "Transformers enabled: {:?}", transformers);
        Self {
            config,
            transformers,
        }
    }
}

#[tokio::main]
async fn main() {
    env_logger::init();

    let config = Config::builder()
        .add_source(
            config::Environment::with_prefix("")
                .prefix_separator("")
                .try_parsing(true)
                .separator("__")
                .list_separator(",")
                .with_list_parse_key("mqtt_topic")
                .with_list_parse_key("mqtt_ignored_topics")
                .with_list_parse_key("esphome_topic_prefixes")
                .with_list_parse_key("hubitat_topic_prefixes"),
        )
        .build()
        .unwrap();
    let app_config = config.try_deserialize().unwrap();
    info!(target: "mqtt-exporter::config", "Loaded config: {app_config:?}");

    let app = App::new(app_config);
    app.prom_task().await;
    app.mqtt_task().await;
}

==> 0.1.2.zip: mqtt-exporter-rs/src/main.rs <==
use backoff::backoff::Backoff;
use config::Config;
use core::str;
use json::JsonValue;
use log::{debug, error, info, warn};
use metrics::gauge;
use metrics_exporter_prometheus::PrometheusBuilder;
use metrics_util::MetricKindMask;
use regex::Regex;
use rumqttc::{
    AsyncClient, ConnAck,
    Event::{self},
    EventLoop, MqttOptions, Packet, Publish, QoS, SubscribeFilter,
};
use serde::Deserialize;
use std::{
    net::{IpAddr, Ipv6Addr},
    process::exit,
    time::{Duration, SystemTime, UNIX_EPOCH},
};

use mqtt_exporter::transforms::*;

struct ConfigDefaults;
impl ConfigDefaults {
    fn mqtt_address() -> String {
        "localhost".to_owned()
    }
    fn mqtt_port() -> u16 {
        1883
    }
    fn mqtt_topic() -> Vec<String> {
        vec!["#".to_owned()]
    }
    fn mqtt_keepalive() -> u64 {
        60
    }
    fn mqtt_client_id() -> String {
        "mqtt-exporter-rs".to_owned()
    }
    fn mqtt_expose_client_id() -> bool {
        true
    }
    fn expose_last_seen() -> bool {
        false
    }
    fn prometheus_address() -> IpAddr {
        IpAddr::V6(Ipv6Addr::from_bits(0))
    }
    fn prometheus_port() -> u16 {
        9000
    }
    fn prometheus_prefix() -> String {
        "mqtt".to_owned()
    }
    fn topic_label() -> String {
        "topic".to_owned()
    }
    fn hubitat_topic_prefixes() -> Vec<String> {
        vec!["hubitat".to_owned()]
    }
    fn zigbee2mqtt_availability() -> bool {
        false
    }
    fn zwave_topic_prefix() -> String {
        "zwave/".to_owned()
    }
}

#[derive(Debug, Deserialize)]
struct AppConfig {
    #[serde(default = "ConfigDefaults::mqtt_address")]
    mqtt_address: String,
    #[serde(default = "ConfigDefaults::mqtt_port")]
    mqtt_port: u16,
    #[serde(default = "ConfigDefaults::mqtt_topic")]
    mqtt_topic: Vec<String>,
    #[serde(default = "ConfigDefaults::mqtt_keepalive")]
    mqtt_keepalive: u64,
    #[serde(default)]
    mqtt_username: Option<String>,
    #[serde(default)]
    mqtt_password: Option<String>,
    #[serde(default = "ConfigDefaults::mqtt_client_id")]
    mqtt_client_id: String,
    #[serde(default = "ConfigDefaults::mqtt_expose_client_id")]
    mqtt_expose_client_id: bool,
    #[serde(default = "ConfigDefaults::expose_last_seen")]
    expose_last_seen: bool,
    #[serde(default)]
    mqtt_ignored_topics: Vec<String>,
    #[serde(with = "serde_regex", default)]
    mqtt_ignored_topics_re: Option<Regex>,
    #[serde(default = "ConfigDefaults::prometheus_address")]
    prometheus_address: IpAddr,
    #[serde(default = "ConfigDefaults::prometheus_port")]
    prometheus_port: u16,
    #[serde(default = "ConfigDefaults::prometheus_prefix")]
    prometheus_prefix: String,
    #[serde(default = "ConfigDefaults::topic_label")]
    topic_label: String,
    #[serde(default)]
    esphome_topic_prefixes: Vec<String>,
    #[serde(default = "ConfigDefaults::hubitat_topic_prefixes")]
    hubitat_topic_prefixes: Vec<String>,
    #[serde(default = "ConfigDefaults::zigbee2mqtt_availability")]
    zigbee2mqtt_availability: bool,
    #[serde(default = "ConfigDefaults::zwave_topic_prefix")]
    zwave_topic_prefix: String,
}
struct App {
    config: AppConfig,
    client: AsyncClient,
    transformers: Vec<Box<dyn Transform>>,
}

fn parse_number(v: &str) -> Option<f64> {
    if let Ok(num) = v.parse::<f64>() {
        Some(num)
    } else if v.eq_ignore_ascii_case("on")
        || v.eq_ignore_ascii_case("true")
        || v.eq_ignore_ascii_case("online")
    {
        Some(1.0)
    } else if v.eq_ignore_ascii_case("off")
        || v.eq_ignore_ascii_case("false")
        || v.eq_ignore_ascii_case("offline")
    {
        Some(0.0)
    } else {
        None
    }
}

impl App {
    fn parse_metric(value: &JsonValue) -> Option<f64> {
        match value {
            JsonValue::Null => None,
            JsonValue::Boolean(v) => {
                if *v {
                    Some(1.0)
                } else {
                    Some(0.0)
                }
            }
            JsonValue::Number(v) => Some((*v).into()),
            JsonValue::Short(v) => parse_number(v.as_str()),
            JsonValue::String(v) => parse_number(v.as_str()),
            _ => None,
        }
    }

    fn parse_metrics(
        &self,
        mut data: JsonValue,
        topic: &str,
        _original_topic: &str,
        metric_path: Vec<String>,
        labels: &Vec<(String, String)>,
    ) {
        match data {
            JsonValue::Array(_) => {
                for (metric, child) in data.members_mut().enumerate() {
                    let mut metric_path = metric_path.clone();
                    metric_path.push(metric.to_string());
                    self.parse_metrics(child.take(), topic, _original_topic, metric_path, labels)
                }
            }
            JsonValue::Object(_) => {
                for (metric, child) in data.entries_mut() {
                    let mut metric_path = metric_path.clone();
                    metric_path.push(metric.to_string());
                    self.parse_metrics(child.take(), topic, _original_topic, metric_path, labels)
                }
            }
            _ => {
                if let Some(v) = Self::parse_metric(&data) {
                    let metric_s = metric_path.join("_");
                    let mut labels = labels.clone();
                    labels.push((self.config.topic_label.to_owned(), topic.to_owned()));
                    debug!(target: "mqtt-exporter::prom", "metric: {metric_s} value: {v} labels: {labels:?}");

                    if self.config.expose_last_seen {
                        let metric_ts = format!("{}_ts", metric_s);
                        let ts_gauge = gauge!(metric_ts, &labels);
                        let ts = SystemTime::now().duration_since(UNIX_EPOCH).unwrap();
                        ts_gauge.set(ts)
                    }
                    let gauge = gauge!(metric_s, &labels);
                    gauge.set(v);
                }
            }
        }
    }

    fn topic_filter(&self, topic: &str) -> bool {
        !self
            .config
            .mqtt_ignored_topics
            .iter()
            .any(|pfx| topic.starts_with(pfx))
            && !self
                .config
                .mqtt_ignored_topics_re
                .as_ref()
                .is_some_and(|re| re.is_match(topic))
    }

    fn deserialize_payload(payload: &[u8]) -> Result<JsonValue, Box<dyn std::error::Error>> {
        Ok(json::parse(str::from_utf8(payload)?)?)
    }

    async fn handle_event(&self, event: &Event) {
        match event {
            Event::Incoming(Packet::Publish(notification)) => self.handle_publish(notification),
            Event::Incoming(Packet::ConnAck(ack)) => self.handle_connack(ack).await,
            // Event::Incoming(Packet::SubAck(ack)) => self.handle_suback(ack),
            Event::Outgoing(_) => {}
            e => debug!(target: "mqtt-exporter::mqtt", "Unhandled event: {:?}", e),
        }
    }

    fn handle_publish(&self, event: &Publish) {
        match Self::deserialize_payload(&event.payload) {
            Ok(parsed) => self.update_metrics(&event.topic, parsed),
            Err(e) => {
                error!(target: "mqtt-exporter::mqtt", "Error deserializing JSON payload: {:?}", e)
            }
        }
    }

    async fn handle_connack(&self, ack: &ConnAck) {
        match ack.code {
            rumqttc::ConnectReturnCode::Success => {
                info!(target: "mqtt-exporter::mqtt", "Successfully connected to MQTT broker");
                self.subscribe().await;
            }
            code => {
                warn!(target: "mqtt-exporter::mqtt", "Non success connection return code {:?}", code)
            }
        }
    }

    fn update_metrics(&self, raw_topic: &str, mut data: JsonValue) {
        let mut labels = Vec::new();

        if !&self.topic_filter(raw_topic) {
            return;
        }
        debug!(target: "mqtt-exporter::mqtt", "topic {} payload {}", raw_topic, data.pretty(2));
        let mut topic = raw_topic.to_owned();

        (topic, data) = self
            .transformers
            .iter()
            .filter(|txfm| txfm.applies_to(raw_topic))
            .fold((topic, data), |(topic, data), txfm| {
                txfm.transform(&topic, data, &mut labels)
            });

        // Should be a transform itself?
        // let topic = topic.replace('/', "_");

        let metric_path = vec![self.config.prometheus_prefix.clone()];
        self.parse_metrics(data, &topic, &topic, metric_path, &labels);
    }

    async fn subscribe(&self) {
        match self
            .client
            .subscribe_many(self.config.mqtt_topic.iter().map(|topic| SubscribeFilter {
                path: topic.to_owned(),
                qos: QoS::AtLeastOnce,
            }))
            .await
        {
            Ok(_) => {
                info!(target: "mqtt-exporter::mqtt", "Successfully subscribed to topics `{}`", &self.config.mqtt_topic.join(" "))
            }
            Err(e) => {
                error!(target: "mqtt-exporter::mqtt", "Failed to subscribe to topics `{}` - {}", &self.config.mqtt_topic.join(" "), e)
            }
        }
    }

    async fn mqtt_task(self, mut eventloop: EventLoop) {
        let mut bo = backoff::ExponentialBackoff::default();

        loop {
            match eventloop.poll().await {
                Ok(event) => {
                    bo.reset();
                    self.handle_event(&event).await
                }
                Err(e) => {
                    error!(target: "mqtt-exporter::mqtt", "Connection error: {}", e);
                    let delay = bo.next_backoff();
                    match delay {
                        Some(delay) => {
                            debug!(target: "mqtt-exporter::mqtt", "Backing off for {}s", delay.as_secs_f32());
                            tokio::time::sleep(delay).await
                        }
                        None => {
                            error!(target: "mqtt-exporter::mqtt", "Connection timed out, giving up");
                            exit(-1)
                        }
                    }
                }
            }
        }
    }
    async fn prom_task(&self) {
        let builder = PrometheusBuilder::new()
            .with_http_listener((self.config.prometheus_address, self.config.prometheus_port))
            .idle_timeout(MetricKindMask::ALL, Some(Duration::from_secs(300)));
        let builder = if self.config.mqtt_expose_client_id {
            builder.add_global_label("client_id", &self.config.mqtt_client_id)
        } else {
            builder
        };
        builder.install().unwrap();
    }

    fn new(config: AppConfig, client: AsyncClient) -> Self {
        let mut transformers: Vec<Box<dyn Transform>> = Vec::new();
        if !config.zwave_topic_prefix.is_empty() {
            transformers.push(Box::new(ZwaveNormalize::new(&config.zwave_topic_prefix)))
        }
        if !config.esphome_topic_prefixes.is_empty() {
            transformers.push(Box::new(EsphomeNormalize::new(
                &config.esphome_topic_prefixes,
            )))
        }
        if !config.hubitat_topic_prefixes.is_empty() && !config.hubitat_topic_prefixes[0].is_empty()
        {
            transformers.push(Box::new(HubitatNormalize::new(
                &config.hubitat_topic_prefixes,
            )))
        }
        if config.zigbee2mqtt_availability {
            transformers.push(Box::new(Zigbee2MqttAvailability::default()))
        }
        debug!(target: "mqtt-exporter::config", "Transformers enabled: {:?}", transformers);

        Self {
            config,
            client,
            transformers,
        }
    }
}

#[tokio::main]
async fn main() {
    env_logger::init();

    let config = Config::builder()
        .add_source(
            config::Environment::with_prefix("")
                .prefix_separator("")
                .try_parsing(true)
                .separator("__")
                .list_separator(",")
                .with_list_parse_key("mqtt_topic")
                .with_list_parse_key("mqtt_ignored_topics")
                .with_list_parse_key("esphome_topic_prefixes")
                .with_list_parse_key("hubitat_topic_prefixes"),
        )
        .build()
        .unwrap();
    let app_config: AppConfig = config.try_deserialize().unwrap();
    info!(target: "mqtt-exporter::config", "Loaded config: {app_config:?}");

    let mut opts = MqttOptions::new(
        &app_config.mqtt_client_id,
        &app_config.mqtt_address,
        app_config.mqtt_port,
    );
    opts.set_keep_alive(Duration::from_secs(app_config.mqtt_keepalive))
        .set_max_packet_size(150000, 150000);

    if let (Some(username), Some(password)) = (&app_config.mqtt_username, &app_config.mqtt_password)
    {
        opts.set_credentials(username, password);
    }

    let (client, mut eventloop) = AsyncClient::new(opts, 10);

    let app = App::new(app_config, client);
    app.prom_task().await;
    app.mqtt_task(eventloop).await;
}
